/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <stdlib.h>
#include <string.h>

#include "voms_apic.h"

extern int retrieve(X509 *, STACK_OF(X509) *, int, char **, int *, char **, int *, char **, int *, char **, char **, int *);
extern char **listjoin(char **, char **, int);
extern char **listadd(char **, char *, int);

typedef int (*cmpf)(struct contactdata *, char *);

static struct contactdata **loadfile(struct contactdata **, char *, uid_t, gid_t, cmpf, char *, int *);
static struct contactdata **loadfile0(struct contactdata **, char *, cmpf, char *, int *);
static int tokenize(char **, unsigned int *, char **);

void 
VOMS_Destroy(struct vomsdata *vd)
{
  int i = 0, j=0;
  struct voms *v;
  struct data *d;
  char *buf;

  if (vd) {
    if (vd->data) {
      while (v = (vd->data)[i++]) {
	if (v->std && (v->type == TYPE_STD)) {
	  j=0;
	  while (d = v->std[j++])
	    free(d);
	  free (v->std);
	}
	if (v->reserved)
	  free(v->reserved);
	free(v);
      }
      free(vd->data);
    }
    if (vd->reserved) {
      i=0;
      while (buf = (vd->reserved)[i++])
	free(buf);
      free(vd->reserved);
    }
    if (vd->workvo)
      free (vd->workvo);
    if (vd->extra_data)
      free (vd->extra_data);

    free(vd->ordering);
    free(vd);
  }
}


struct voms *
VOMS_DefaultData(struct vomsdata *vd, int *error)
{
  if (!error)
    return NULL;

  *error = VERR_NONE;

  if (!vd || ! vd->data || ! ((vd->data)[0])) {
    *error = VERR_NOINIT;
    return NULL;
  }

  return  ((vd->data)[0]);
}

struct vomsdata *
VOMS_Init(char *v, char *c)
{
  int result = 1;
  struct vomsdata *tmp;
  DIR *vdir, *cdir;

  if (!v)
    if (!(v = getenv("X509_VOMS_DIR")))
      v = "/etc/grid-security/vomsdir";

  if (!c)
    if (!(c = getenv("X509_CERT_DIR")))
      c = "/etc/grid-security/certificates";

  vdir = opendir(v);
  cdir = opendir(c);

  if (!vdir || !cdir)
    result = 0;
    
  if (cdir)
    (void)closedir(cdir);
  if (vdir)
    (void)closedir(vdir);

  if (!result)
    return NULL;

  if (tmp = (struct vomsdata *)malloc(sizeof(struct vomsdata))) {
    tmp->cdir = c;
    tmp->vdir = v;
    tmp->extra_data = tmp->workvo = NULL;
    tmp->data = NULL;
    tmp->reserved = NULL;
    tmp->reserved2=0;
    tmp->ordering=NULL;
  }
  return tmp;

}

extern int verify(char *message, int len, struct vomsdata *voms, char *subject, char *ca, int *error);

int
VOMS_SetLifetime(int length, struct vomsdata *vd, int *error)
{
  if (!error)
    return 0;
  if (!vd) {
    *error= VERR_NOINIT;
    return 0;
  }

  vd->reserved2 = length;

  return 1;
}

int 
VOMS_ResetOrder(struct vomsdata *vd, int *error)
{

  if (!error)
    return 0;
  if (!vd) {
    *error = VERR_NOINIT;
    return 0;
  }

  free(vd->ordering);
  vd->ordering = NULL;
  return 1;
}

int 
VOMS_Ordering(char *order, struct vomsdata *vd, int *error)
{
  int length = 0;
  char *tmp;

  if (!error)
    return 0;
  if (!vd) {
    *error = VERR_NOINIT;
    return 0;
  }
  if (!order) {
    *error = VERR_PARAM;
    return 0;
  }
  if (vd->ordering != NULL)
    length = strlen(vd->ordering);

  length += strlen(order) + 3;
  if (tmp = (char *)malloc(length)) {
    if (vd->ordering != NULL)
      strcpy(tmp, vd->ordering);
    else
      strcpy(tmp, ",");
    strcat(tmp, order);
    strcat(tmp, ",");
    free(vd->ordering);
    vd->ordering=tmp;
    return 1;
  }

  *error = VERR_MEM;
  return 0;
}

int 
VOMS_Contact(char *hostname, int port, char *servsubject, char *command, struct vomsdata *vd, int *error)
{
  char *buffer;
  int buflen;
  char *subject, *ca;
  int res = 0;
  int length = 10;
  char *tmp;
  int i =0;

  subject = ca = NULL;

  if (!error)
    return 0;
  if (!vd) {
    *error = VERR_NOINIT;
    return 0;
  }

  *error = VERR_NOSOCKET;

  length += strlen(command) + 2;
  if (vd->ordering)
    length += strlen(vd->ordering);

  if (tmp = (char *)malloc(length)) {
    sprintf(tmp, "%ld,", vd->reserved2);
    if (vd->ordering) {
      strcat(tmp, vd->ordering);
      strcat(tmp, ",");
    }
    strcat(tmp, command);
    if (net_connect(hostname, port, servsubject, tmp, strlen(tmp), &buffer, &buflen, &subject, &ca)) {
      vd->extra_data = vd->workvo = NULL;
      vd->extralen = vd->volen = 0;
      while (buffer[0] == '1') {
	while ((buffer[i] != '\1') && (i < buflen))
	  i++;
	if (i == buflen) {
	  *error = VERR_FORMAT;
	  free(buffer);
	  free(subject);
	  free(ca);
	  return 0;
	}
      }
      while (buffer[0] == '2') {
	*error = VERR_NODATA;
	free(buffer);
	free(subject);
	free(ca);
	return 0;
      }
      while (buffer[0] == '0') {
	while ((buffer[i] != '\1') && (i < buflen))
	  i++;
	if (i == buflen) {
	  *error = VERR_FORMAT;
	  free(buffer);
	  free(subject);
	  free(ca);
	  return 0;
	}
	else 
	  *error = VERR_ORDER;
      }
      if (subject && ca)
	res = verify(buffer+i, buflen-i, vd, subject, ca, error);
      else
	free(buffer);
      free(subject);
      free(ca);
    }

    free(tmp);
    return res;
  }
  *error = VERR_MEM;
  return 0;
}

int 
VOMS_Retrieve(X509 *cert, STACK_OF(X509) *chain, int how, struct vomsdata *vd, int *error)
{
  char *buffer;
  char *vo;
  char *file;
  int buflen, volen, filelen;
  char *subject,  *ca;
  int res = 0;

  subject = ca = NULL;

  buffer = vo = file = NULL;

  if (!error)
    return 0;
  if (!vd) {
    *error = VERR_NOINIT;
    return 0;
  }

  if (retrieve(cert, chain, how, &buffer, &buflen, &vo, &volen, &file, &filelen, &subject, &ca, error)) {
    vd->workvo = vo;
    vd->extra_data = file;
    vd->volen = volen;
    vd->extralen = filelen;
    if (subject && ca)
      res = verify(buffer, buflen, vd, subject, ca, error);
  }

  free(subject);
  free(ca);
  
  return res;
}


int VOMS_Import(char *buffer, int buflen, struct vomsdata *vd, int *error)
{
  if (!error)
    return 0;
  if (!vd) {
    *error = VERR_NOINIT;
    return 0;
  }

  return verify(buffer, buflen, vd, NULL, NULL, error);
}

static int check_alias(struct contactdata *cd, char *alias)
{
  return strcmp(cd->nick, alias);
}

static int check_vo(struct contactdata *cd, char *vo)
{
  return strcmp(cd->vo, vo);
}

struct contactdata **VOMS_FindByAlias(struct vomsdata *vd, char *alias, char *system, char *user, int *error)
{
  struct contactdata **list = NULL;
  *error = VERR_NONE;

  if (!system)
    system="/opt/edg/etc/vomses";
  if (!user) {
    char *uc   = getenv("VOMS_USERCONF");
    char *home = getenv("HOME");

    if (uc)
      user = uc;
    else if (home) {
      if (user = alloca(strlen(home) + 15)) {
	strcpy(user, home);
	strcat(user, "/.edg/vomses/");
      }
      else
	return NULL;
    }
    else
      user = "~/.edg/vomses";
  }
  list = loadfile(NULL, system, 0, 0, check_alias, alias, error);
  if (*error == VERR_NONE)
    return loadfile(list, user, getuid(), getgid(), check_alias, alias, error);
  else {
    if (list)
      VOMS_DeleteContacts(list);
    return NULL;
  }
}

struct contactdata **VOMS_FindByVO(struct vomsdata *vd, char *alias, char *system, char *user, int *error)
{
  struct contactdata **list = NULL;

  if (!system)
    system="/opt/edg/etc/vomses";
  if (!user) {
    char *uc   = getenv("VOMS_USERCONF");
    char *home = getenv("HOME");

    if (uc)
      user = uc;
    else if (home) {
      if (user = alloca(strlen(home) + 15)) {
	strcpy(user, home);
	strcat(user, "/.edg/vomses/");
      }
      else
	return NULL;
    }
    else
      user = "~/.edg/vomses";
  }
  list = loadfile(NULL, system, 0, 0, check_vo, alias, error);
  if (*error == VERR_NONE)
    return loadfile(list, user, getuid(), getgid(), check_vo, alias, error);
  else {
    if (list)
      VOMS_DeleteContacts(list);
    return NULL;
  }
}

void VOMS_DeleteContacts(struct contactdata **list)
{
  int i=0;

  if (!list)
    return;

  while (list[i]) {
    free(((struct contactdata *)(list[i]))->reserved);
    free(list[i]);
    i++;
  }
  free(list);
}

static int
tokenize(char **str, unsigned int *start, char **value)
{
  char *beg, *end;
  beg = strchr(*str, '"');
  if (beg) {
    *value = ++beg;
    end = strchr(beg, '"');
    if (end) {
      *end++='\0';
      *str = end;
      return 1;
    }
  }
  return 0;
}

static struct contactdata **loadfile(struct contactdata **prev, char *file, uid_t uid, gid_t gid, cmpf check, char *param, int *error)
{
  struct stat stats;
  DIR *dp = NULL;

  if (!file)
    return 0;

  if (stat(file, &stats) == -1 || stats.st_uid != uid ||
      stats.st_gid != gid ||
      !(stats.st_mode == (S_IFREG | S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) ||
       stats.st_mode == (S_IFDIR |
			 S_IRUSR | S_IWUSR | S_IXUSR |
			 S_IRGRP | S_IXGRP |
			 S_IROTH | S_IXOTH))) {
    return 0;
  }

  if (stats.st_mode & S_IFREG)
    return loadfile0(prev, file, check, param, error);
  else {
    struct dirent *de;

    dp = opendir(file);
    if (dp) {
      while (de = readdir(dp)) {
	char *name = de->d_name;
	if (name && (strcmp(name, ".") != 0) && (strcmp(name, "..") != 0)) {
	  char *fname = alloca(strlen(file)+strlen(name)+2);
	  if (fname) {
	    struct contactdata **data;

	    strcpy(fname, file);
	    strcat(fname, "/");
	    strcat(fname, name);
	    data = loadfile(prev, fname, uid, gid, check, param, error);
	    if (data)
	      prev = data;
	  }
	  else {
	    *error = VERR_MEM;
	    goto err;
	  }
	}
      }
      closedir(dp);
      return prev;
    }
    else {
      *error = VERR_DIR;
      goto err;
    }
  }

  return prev;

 err:
  if (dp)
    closedir(dp);
  return NULL;
}

static struct contactdata **loadfile0(struct contactdata **prev, char *file, cmpf check, char *param, int *error)
{
  struct contactdata tmpdata;
  struct contactdata *data = NULL;
  struct contactdata **current = NULL;
  struct contactdata **result;
  FILE *f;

  if (!file)
    return 0;

  /* Opens the file */
  if (!(f = fopen(file, "r")))
    return 0;

  /* Load the file */

  while (!feof(f)) {
    char *line = NULL;
    int t=0;

    if (getline(&line, &t, f) != -1) {
      unsigned int start = 0;
      char *port;
      int ok = 1;

      tmpdata.reserved = line;
      ok &= tokenize(&line, &start, &(tmpdata.nick));
      ok &= tokenize(&line, &start, &(tmpdata.host));
      ok &= tokenize(&line, &start, &port);
      ok &= tokenize(&line, &start, &(tmpdata.contact));
      ok &= tokenize(&line, &start, &(tmpdata.vo));
      tmpdata.port=atoi(port);

      line = NULL;
      t=0;

      if (!ok) {
	free(tmpdata.reserved);
	goto err;
      }
      else if (check(&tmpdata, param)) {
	free(tmpdata.reserved);
	continue;
      }
      else {
	struct contactdata **tmp  = NULL;

	data = NULL;
	if (data = (struct contactdata *)malloc(sizeof(struct contactdata))) {
	  memcpy(data, &tmpdata, sizeof(struct contactdata));
	  
	  tmp = (struct contactdata **)listadd((char **)current, (char *)data, sizeof(struct contactdata *));
	  if (tmp) {
	    current = tmp;
	    continue;
	  }
	}
	*error=VERR_MEM;
	free(data);
	free(tmpdata.reserved);
	goto err;
      }
    }
    else {
      if (!feof(f)) {
	if (line)
	  free(line);
	*error = VERR_MEM;
	goto err;
      }
      else if (line)
	free(line);
    }
  }

  /*
   * No error and file completely read.
   */

  if (f) {
    fclose(f);
    f = NULL;
  }

  result = (struct contactdata **)listjoin((char **)prev, (char **)current, sizeof(struct contactdata *));

  if (result) {
    free(current);
    return result;
  }

 err:
  if (f)
    fclose(f);
  if (current)
    VOMS_DeleteContacts(current);
  return NULL;
}


int getMajorVersionNumber(void) {return 1;}
int getMinorVersionNumber(void) {return 4;}
int getPatchVersionNumber(void) {return 0;}

