#ifndef __SE_FILE_H__
#define __SE_FILE_H__

#include "../../std.h"

#include <string>
#include <list>
#include <iostream>

#include "se_ns.h"
#include "../../misc/condition.h"
#include "../../auth/auth.h"
#include "../../misc/safelist.h"

#include "gacl.h"

/*
class SEUserDescription {
 public:
  string subject;
  list<string> group; // any information used to group users presented in one line
};
*/

// Access control
#define FILE_ACC_NONE        GACL_PERM_NONE
#define FILE_ACC_LIST        GACL_PERM_LIST
#define FILE_ACC_READ        GACL_PERM_READ
#define FILE_ACC_WRITE       GACL_PERM_WRITE
#define FILE_ACC_DELETE      GACL_PERM_WRITE
#define FILE_ACC_READ_META   GACL_PERM_ADMIN
#define FILE_ACC_WRITE_META  GACL_PERM_ADMIN
#define FILE_ACC_ALL         0xFFFF 


#define FILE_STATE_ACCEPTED       0    // File accepted at SE (initial state)
#define FILE_STATE_COLLECTING     1    // File is not physically complete
#define FILE_STATE_COMPLETE       2    // File is physically complete
#define FILE_STATE_DELETING       3    // File is being deleted
#define FILE_STATE_MAX            4     
#define FILE_STATE_UNCHANGED      4

#define REG_STATE_LOCAL           0
#define REG_STATE_REGISTERING     1  // File claimed - registration in progress
#define REG_STATE_ANNOUNCED       2
#define REG_STATE_UNREGISTERING   3  // File claimed - registration in progress
#define REG_STATE_MAX             4
#define REG_STATE_UNCHANGED       4

extern const char* file_state_str[];
extern const char* reg_state_str[];

#define FILE_STATE_ACCEPTED_STR      file_state_str[FILE_STATE_ACCEPTED]
#define FILE_STATE_COLLECTING_STR    file_state_str[FILE_STATE_COLLECTING
#define FILE_STATE_COMPLETE_STR      file_state_str[FILE_STATE_COMPLETE]
#define FILE_STATE_DELETING_STR      file_state_str[FILE_STATE_DELETING]

#define REG_STATE_LOCAL_STR          reg_state_str[REG_STATE_LOCAL]
#define REG_STATE_REGISTERING_STR    reg_state_str[REG_STATE_REGISTERING]
#define REG_STATE_ANNOUNCED_STR      reg_state_str[REG_STATE_ANNOUNCED]
#define REG_STATE_UNREGISTERING_STR  reg_state_str[REG_STATE_UNREGISTERING]



class SEAuthorization {

};

// File matadata
class SEAttributes {
 private:
  bool valid;
  uint64_t size_i;
  bool size_b;
  string id_i;
  string creator_i;
  string checksum_i;
  bool checksum_b;
  struct tm created_i;
  bool created_b;
 public:
  operator bool(void) { return valid; };
  bool enough(void);
  bool complete(void);
//  bool full(void);
  bool operator!(void) { return !valid; };
  SEAttributes(void):valid(false) { };
  SEAttributes(const char* id,uint64_t size,const AuthUser user):id_i(id),size_b(true),size_i(size),valid(true) {
    checksum_b=false;
    created_b=false;
    creator_i=user.DN();
  };
  SEAttributes(const SEAttributes& attr):valid(true) {
    id_i=attr.id_i;
    size_i=attr.size_i; size_b=true;
    if((checksum_b=attr.checksum_b)) checksum_i=attr.checksum_i;
    if((created_b=attr.created_b)) memcpy(&created_i,&(attr.created_i),sizeof(created_i));
    creator_i=attr.creator_i;
  };
  SEAttributes(const char* fname);
  void operator=(const SEAttributes& attr) {
    id_i=attr.id_i;
    size_i=attr.size_i;
    checksum_b=attr.checksum_b;
    checksum_i=attr.checksum_i;
    created_b=attr.created_b;
    memcpy(&created_i,&(attr.created_i),sizeof(created_i));
    creator_i=attr.creator_i;
  };
  int read(const char* fname);
  int write(const char* fname);
  uint64_t size(void) const { return size_i; };
  void size(uint64_t s) { size_i=s; };
  const char* id(void) const { return id_i.c_str(); };
  void id(const char* i) { id_i=i; };
  const string& creator(void) const { return creator_i; };
  const string& checksum(void) const { return checksum_i; };
  void checksum(const string& m) { checksum_i=m; checksum_b=true; };
  bool checksum_available(void) const { return checksum_b; };
  bool checksum_compare(const string& m) { return (checksum_i==m); };
  const struct tm* created(void) const { return &created_i; };
  void created(const struct tm* c) { memcpy(&created_i,c,sizeof(created_i)); created_b=true; };
  void created(const time_t* c);
  void created(const char* c);
  void created(const string& c) { created(c.c_str()); };
  bool created_available(void) const { return created_b; };
  bool created_compare(const char* c);
  // void md5_string(string &s);
};

typedef struct {
  uint64_t start;
  uint64_t end;
} SEFileRange;
#define MAX_SEFILE_RANGES 100

class SEState {
 friend ostream& operator<<(ostream& o,const SEState &s);
 private:
  int file_;
  int reg_;
  time_t file_changed;
  time_t reg_changed;
 public:
  SEState(void):file_(FILE_STATE_ACCEPTED),reg_(REG_STATE_LOCAL) {
    file_changed=time(NULL);
    reg_changed=time(NULL);
  };
  SEState(int f,int r):file_(f),reg_(r) {
    file_changed=time(NULL);
    reg_changed=time(NULL);
  };
  ~SEState() { };
  int file(void) const { return file_; };
  int reg(void) const { return reg_; };
  int file(int f) {
    if((f>=FILE_STATE_UNCHANGED) || (f<0)) return file_;
    return file_=f;
  };
  int reg(int r) {
    if((r>=REG_STATE_UNCHANGED) || (r<0)) return reg_;
    return reg_=r;
  };
  bool set(const char* key,const char* value);
};

ostream& operator<<(ostream& o,const SEState &s);

class SEFileHandle;

// File itself
class SEFile: public SEAuthorization, public SEAttributes {
 friend class SEFileHandle;
 private:
  bool valid;

  string path;          // Path to directory to store file
  const char* name;     // Name on file system

  SEFileRange* ranges;  // Ranges available (file is complete if NULL)
  LockSimple lock;      // use to atomically perform modifications (used 
                        // only from outside of that class)
  LockSimple lock_;     // Lock for internal operations
  int read_count;       // how many time it is open for reading
  int write_count;      // how many times it is open for writing
  int usage_count;      // not used yet
  int file_handle;      // Handle for file access

  SEState state_;       // State of file and registration (is handled 
                        // outside of this class yet)

 public:
  operator bool(void) { return valid; };
  SEFile(const char* dirpath,const SEAttributes&);
  SEFile(const char *fname);
  ~SEFile(void);
  bool check_id(const char* id_) const { return (strcmp(id(),id_)==0); };
  bool check_name(const char* name_) const { return (strcmp(name,name_)==0); };
  int open(bool for_read);
  void close(bool for_read);
  uint64_t write(void* buf,uint64_t offset,uint64_t size);
  uint64_t read(void* buf,uint64_t offset,uint64_t size);
  int check_acl(const AuthUser &user);
  int write_acl(const AuthUser &user,const char* acl);
  int read_acl(const AuthUser &user,string& acl);
  int write_attr(void);
  bool state_file(int f);
  bool state_reg(int r);
  int state_file(void) { return state_.file(); };
  int state_reg(void) { return state_.reg(); };
  void acquire(void) { lock.block(); };     // lock for atomic operation
  void release(void) { lock.unblock(); };   // release after atomic operation
  void destroy(void);
};

class SENameServer;

#define registration_immediate 1
#define registration_retry 2
class SEFiles {
 private:
/*
  typedef struct {
    SEFile* f;
    int u;         // usage counter
  } SEFile_p;
  SEFile_p* files;     // Files handled by this entry point
*/
  SafeList<SEFile> files;
//  uint32_t nfiles;
  SENameServer* ns;   // Server for registrations to be sent
  bool valid;
  string path; // directory path to store new files
  LockSimple lock;
  SEFile* find_by_id(const char* id); 
  int reg_type;
 public:
/*
  class iterator { // to go through list of files in safe way
   private:
    SEFiles* f;
    int n;
   public:
    iterator(void) { n=0; f=NULL; };
    iterator(const iterator &i) { f=i.f; n=i.n; };
    void operator ++(void) { if(p) p++; if(*p == NULL) p=NULL; }
//    void operator ++(iterator& i) { if(i.p) p++; if(*i.p == NULL) i.p=NULL; }
    bool operator ==(const iterator& i) const { return (p == i.p); };
    SEFile* operator ->(void) const { return (*p); };
    SEFile& operator *(void) const { return (*(*p)); };
    bool operator==(const iterator &i) { return p == i.p; };
    bool operator!=(const iterator &i) { return p != i.p; };
  };
  iterator begin() { iterator i; i.n=0; i.f=this; return i; };
  iterator end() { iterator i; i.n=-1; i.f=this; return i; };
*/
  typedef SafeList<SEFile>::iterator iterator;
  iterator begin() { return files.begin(); };
  iterator end() { return files.end(); };
  void acquire(void) { lock.block(); };
  void release(void) { lock.unblock(); };
  SEFiles();
  SEFiles(const char* dirpath);
  ~SEFiles();
  operator bool(void) { return valid; };
  SEFile* add(SEFile& file);
  bool remove(iterator& file);
  bool remove(SEFile& file);
  SEFile* get(const char* id); 
  SENameServer* NS(void) { return ns; };
  void NS(SENameServer* ns_,int reg_) {
    if(ns) delete ns;
    ns=ns_; reg_type=reg_;
  };
  void NS(const char* ns_cfg,const char* se_url,int reg_) {
    if(ns) delete ns;
    ns=create_ns(ns_cfg,se_url);
    reg_type=reg_;
  };
  const char* Path(void) { return path.c_str(); };
  int Register(void);
  int Unregister(void);
  bool register_immediately(void) {
    return (reg_type & registration_immediate)?true:false;
  };
  bool register_retry(void) {
    return (reg_type & registration_retry)?true:false;
  };
};

class SEFileHandle {
 private:
  SEFile& file;
  uint64_t offset;
  bool open_read;
  bool active;
 public:
  SEFileHandle(SEFile& f,uint64_t offset,bool open_read);
  ~SEFileHandle(void);
  static SEFileHandle* open(const char* id,uint64_t offset,bool for_read,SEFiles& files);
  static SEFileHandle* open(const char* id,bool for_read,SEFiles& files) { return open(id,0,for_read,files); };
  static SEFileHandle* create(const char* id,bool for_read,SEFiles& files);
  uint64_t write(void* buf,uint64_t size);
  uint64_t read(void* buf,uint64_t size);
  uint64_t tell(void) { return offset; };
  void seek(uint64_t o) { offset=o; };
  void close(void) { if(active) { active=false; file.close(open_read); }; };
  uint64_t size(void) { return file.size(); };
  bool full(void) { return (file.ranges==NULL); };
  operator bool(void) { return active; };
  bool operator!(void) { return !active; };
  SEFile& instance(void) { return file; };
};

#endif // __SE_FILE_H__
